/* ++++++++++
	driver.c
	Gameport / Joystick device driver
	for the
	Creative SBLive
	Creative Audigy
	Creative Audigy2
	
	Uses the generic gameport module.
	
	Written by: Greg Crain   gsc70@comcast.net
	Michigan, USA  ....Go Red Wings......

Ver History:
    12-2002 0.1 didn't work
    12-2002 0.2 fixed a typo, and added SBLive and Audigy2.  Works now.
	03-2003 0.3 cleaned up the code a bit for release 
+++++ */

/* Note:  I never got the MIRROR_IO working correctly.  It doesn't really need 
to be used anyway.   SO don't use it as an example. */

#include <KernelExport.h>
#include <Drivers.h>
#include <Errors.h>
#include <KernelExport.h>
#include <PCI.h>
#include "joystick_driver.h"

#define VENDOR_ID_CREATIVE 0x1102
#define SBLIVE_ID 0x7002
#define AUDIGY_ID 0x7003

#define NUM_CARDS 3
#define DRIVER_NAME "Audigy Joystick Port"
#define MIRROR_IO 0
#define MIRROR_PORT_ADDR 0x201
/* Do we want to mirror the io address to the common location for gameport
or leave it at the base address of the pci config */
#define DEBUG 0

extern device_hooks joy_hooks;
/*=-=-=-=-=-=-=-=-=--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/
/*   Global Variables										*/
/*=-=-=-=-=-=-=-=-=--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/
typedef struct _joy_dev {
	void *		driver;
	char		name1[64];
} joy_dev;

typedef struct _audigy_pci_dev {
//	char		name[DEVNAME];	/* used for resources */
	pci_info	info;
	joy_dev		joy;
} audigy_pci_dev;

unsigned int base_addr;
pci_module_info *pci_bus;
static char pci_name[] = B_PCI_MODULE_NAME;
static pci_module_info	*pci;
int num_cards;
int num_names;
generic_gameport_module *gameport;
static char gameport_name[] = "generic/gameport/v1";
audigy_pci_dev card;
char * names[NUM_CARDS*7+1];
pci_info info;
/*=-=-=-=-=-=-=-=-=--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/

/* ----------
	init_hardware - called once the first time the driver is loaded
----- */
status_t
init_hardware (void)
{
 	int ix=0;
	status_t err = ENODEV;
	num_cards = 0;

#if DEBUG	
	bool former = set_dprintf_enabled ( true );
	dprintf("init_hardware - %s\n",DRIVER_NAME);
#endif

	if (get_module(pci_name, (module_info **)&pci))
		return ENOSYS;

	while ((*pci->get_nth_pci_info)(ix, &info) == B_OK) {
		if (info.vendor_id == VENDOR_ID_CREATIVE &&
			( info.device_id == SBLIVE_ID
			|| info.device_id == AUDIGY_ID)) {
			err = B_OK;
			base_addr = info.u.h0.base_registers[0];
			num_cards = num_cards +1;
#if DEBUG
		    dprintf("found Audigy Joystick Port in slot - %x \n",ix);
		    dprintf("init_joy: base addr 0x%x \n",base_addr);
#endif
		}
		ix++;
	}
    
	put_module(pci_name);

//    dprintf("pci name - %s\n",DRIVER_NAME);
//	dprintf("number of Creative gameports: %i",num_cards); 
	
	return B_OK;
}


/* ----------
	init_driver - optional function - called every time the driver
	is loaded.
----- */
status_t
init_driver (void)
{
 	int ix=0;
	status_t err = ENODEV;
	char * name = "Audigy/1";
	num_cards = 0;

#if DEBUG
	dprintf("init_driver - %s\n",DRIVER_NAME);
#endif 

	if (get_module(pci_name, (module_info **)&pci))
		return ENOSYS;

	if (get_module(gameport_name, (module_info **) &gameport)) {
		put_module(pci_name);
		return ENOSYS;
	}
	while ((*pci->get_nth_pci_info)(ix, &info) == B_OK) {
		if (info.vendor_id == VENDOR_ID_CREATIVE &&
			( info.device_id == SBLIVE_ID
			|| info.device_id == AUDIGY_ID)) {
			err = B_OK;
			base_addr = info.u.h0.base_registers[0];
#if DEBUG
		    dprintf("found Audigy Joystick Port in slot - %x \n",ix);
			dprintf("number of Creative gameports: %i",num_cards); 
		    dprintf("init_joy: base addr 0x%x \n",base_addr);
#endif
		}
		ix++;
	}

#if MIRROR_IO
(*pci->write_pci_config) (
						info.bus , 					/* bus number */
						info.device,				/* device # on bus */
						info.function, 				/* function # in device */
						PCI_base_registers,			/* offset in configuration space */
						 4, 						/* # bytes to write (1, 2 or 4) */
						MIRROR_PORT_ADDR			/* value to write */
					);
#endif	
#if DEBUG
	dprintf("Audigy-joy info.bus = 0x%x \n",info.bus);
	dprintf("Audigy-joy info.device = 0x%x\n",info.device);
	dprintf("Audigy-joy info.function = 0x%x\n",info.function);
	dprintf("Audigy-joy: IO ports mirrored at 0x%03x\n", MIRROR_PORT_ADDR);
#endif    

/* Create a device module at Base Addr or Mirror to standard joystick address*/
#if MIRROR_IO
    if ((*gameport->create_device)(MIRROR_PORT_ADDR, &card.joy.driver) <B_OK)
#else
    if ((*gameport->create_device)(base_addr, &card.joy.driver) <B_OK)
#endif
      {
       dprintf("Audigy joystick - Error creating device\n");
       (*gameport->delete_device)(card.joy.driver);
      }

//	else
//	 dprintf("Audigy joystick - module device created\n");	

    sprintf(card.joy.name1, "joystick/%s", name);
	names[num_names++] = card.joy.name1;
#if DEBUG
	dprintf("Joystick name:  %s\n",card.joy.name1);
#endif

	return B_OK;
}


/* ----------
	uninit_driver - optional function - called every time the driver
	is unloaded
----- */
void
uninit_driver (void)
{
#if DEBUG
	dprintf ("uninit_joystick driver: Audigy\n");
#endif
/* Remove the device that was created, and put module back */
	(*gameport->delete_device)(card.joy.driver);
	put_module(gameport_name);
	put_module(pci_name);
#if MIRROR_IO
	(*pci->write_pci_config) (
						info.bus , 			/* bus number */
						info.device,		/* device # on bus */
						info.function, 		/* function # in device */
						PCI_base_registers		,		/* offset in configuration space */
						 4, 				/* # bytes to write (1, 2 or 4) */
						base_addr 			/* value to write */
					);
#endif
}

	
/* ----------
	joy_open - handle open() calls
----- */

static status_t
joy_open (const char *name, uint32 flags, void** cookie)
{
	int ix;
	int offset = -1;
	*cookie = NULL;

#if DEBUG
	dprintf("Audigy: joy_open()\n");
    dprintf("Joystick open looking for name: %s \n", name);
    dprintf("Joystick open, Audigy name: %s \n",card.joy.name1);
#endif
	if (!strcmp(name, card.joy.name1)) {
			offset = 0;
		}
	if (offset < 0) {
	   dprintf("Audigy: joy_open problem offset<0 ENODEV\n");
		return ENODEV;
	}

	return (*gameport->open_hook)(card.joy.driver, flags, cookie);

}


/* ----------
	joy_read - handle read() calls
----- */

static status_t
joy_read (void* cookie, off_t position, void *buffer, size_t* num_bytes)
{
#if DEBUG
	dprintf("joy: read\n");
#endif
	return (*gameport->read_hook)(cookie, position, buffer, num_bytes);
}


/* ----------
	joy_write - handle write() calls
----- */

static status_t
joy_write (void* cookie, off_t position, const void* buffer, size_t* num_bytes)
{
#if DEBUG
    dprintf("joy: write\n");
#endif
	return (*gameport->write_hook)(cookie, position, buffer, num_bytes);
}


/* ----------
	joy_control - handle ioctl calls
----- */

static status_t
joy_control (void* cookie, uint32 op, void* arg, size_t len)
{
#if DEBUG
    dprintf("joy: control\n");   
#endif
	return (*gameport->control_hook)(cookie, op, arg, len);
}


/* ----------
	joy_close - handle close() calls
----- */

static status_t
joy_close (void* cookie)
{
#if DEBUG
	dprintf("joy: close\n");
#endif
	return (*gameport->close_hook)(cookie);	
}


/* -----
	joy_free - called after the last device is closed, and after
	all i/o is complete.
----- */
static status_t
joy_free (void* cookie)
{
	return (*gameport->free_hook)(cookie);	
}


/* -----
	null-terminated array of device names supported by this driver
----- */

static const char *ad_device_name[] = {
	"joystick/Audigy/1",
	NULL
};

 static const char *sb_device_name[] = {
	"joystick/SBLive/1",
	NULL
}; 

/* -----
	function pointers for the device hooks entry points
----- */

device_hooks my_device_hooks = {
	joy_open, 			/* -> open entry point */
	joy_close, 			/* -> close entry point */
	joy_free,			/* -> free cookie */
	joy_control, 		/* -> control entry point */
	joy_read,			/* -> read entry point */
	joy_write			/* -> write entry point */
};

/* ----------
	publish_devices - return a null-terminated array of devices
	supported by this driver.
----- */

const char**
publish_devices()
{
#if DEBUG
   dprintf ("joystick - publish_devices\n");
#endif

   if (info.device_id == AUDIGY_ID)
		return ad_device_name;
	else
		return sb_device_name;
}

/* ----------
	find_device - return ptr to device hooks structure for a
	given device name
----- */

device_hooks*
find_device(const char* name)
{
#if DEBUG
   dprintf("joystick - find devices\n");
#endif
	return &my_device_hooks;
}

